import pytest
from freezegun import freeze_time
from aiohttp import ClientSession
from pyaffalddk import GarbageCollection
from pathlib import Path
import pickle
import json
import os


CI = os.getenv("CI") == "true"
skip_in_ci = pytest.mark.skipif(CI, reason="Skipped in CI environment")
UPDATE = False
ADDRESS_LIST_KEYS = ['id', 'fullname']

datadir = Path(__file__).parent/'data'
kbh_ics_data = (datadir/'kbh_ics.data').read_text()
odense_ics_data = (datadir/'odense_ics.data').read_text()
aalborg_data_gh = json.loads((datadir/'Aalborg_gh.data').read_text())
aarhus_data = json.loads((datadir/'Aarhus.data').read_text())
koege_data = json.loads((datadir/'Koege.data').read_text())
affaldonline_data = json.loads((datadir/'affaldonline.data').read_text())
vestfor_data = pickle.loads((datadir/'vestfor.data').read_bytes())
openexp_data = json.loads((datadir/'openexp.data').read_text())
openexplive_data = json.loads((datadir/'openexplive.data').read_text())
renodjurs_data = json.loads((datadir/'renodjurs.data').read_text())
provas_data = json.loads((datadir/'provas.data').read_text())

FREEZE_TIME = "2025-04-25"
compare_file = (datadir/'compare_data.p')


def update_and_compare(name, actual_data, update=False, debug=False):
    compare_data = pickle.load(compare_file.open('rb'))
    if update:
        compare_data[name] = actual_data
        pickle.dump(compare_data, compare_file.open('wb'))
    if debug and actual_data != compare_data[name]:
        print(actual_data.keys())
        print(compare_data[name].keys())
    assert actual_data == compare_data[name]


async def assert_add_list(gc, address_list):
    add = await gc.get_address(address_list[0])
    print(add.address_id, add.address)
    assert list(gc._api.address_list[address_list[0]].keys()) == ADDRESS_LIST_KEYS


@pytest.mark.asyncio
@freeze_time("2025-05-25")
async def test_OpenExpLive(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Frederiksberg', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Frederiksberg_70984', 'address_id': '70984',
                'kommunenavn': 'Frederiksberg', 'address': 'Smallegade 1'}
            if not CI:
                address_list = await gc.get_address_list('2000', 'Smallegade', '1')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('2000', 'Smallegade', '')
                assert len(address_list) == 80
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('2000', 'Smallegade', '2')
                assert len(address_list) == 9

            async def get_data(*args, **kwargs):
                return openexplive_data
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Frederiksberg', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time("2025-05-20")
async def test_OpenExp(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Viborg', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Viborg_67580', 'address_id': '67580',
                'kommunenavn': 'Viborg', 'address': 'Prinsens alle 5'}
            if not CI:
                address_list = await gc.get_address_list('8800', 'Prinsens Alle', '5')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('8800', 'Vesterled', '')
                assert len(address_list) == 52
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('8800', 'Vesterled', '4')
                assert len(address_list) == 9

            async def get_data(*args, **kwargs):
                return openexp_data
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Viborg', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time("2025-05-20")
async def test_Affaldonline(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Vejle', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Vejle_1261533|490691026|0', 'address_id': '1261533|490691026|0',
                'kommunenavn': 'Vejle', 'address': 'Klostergade 2a'}
            if not CI:
                address_list = await gc.get_address_list('7100', 'Klostergade', '2A')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('7100', 'Vestbanevej', '')
                assert len(address_list) == 52
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('7100', 'Vestbanevej', '1')
                assert len(address_list) == 12

            async def get_data(*args, **kwargs):
                return affaldonline_data
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Vejle', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time("2025-05-04")
async def test_PerfectWaste(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Køge', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Køge_27768', 'address_id': '27768',
                'kommunenavn': 'Køge', 'address': 'Torvet 1'}
            if not CI:
                address_list = await gc.get_address_list('4600', 'Torvet', '1')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('4600', 'Oksbølvej', '')
                assert len(address_list) == 47
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('4600', 'Oksbølvej', '1')
                assert len(address_list) == 11

            async def get_data(*args, **kwargs):
                return koege_data["result"]
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Koege', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time("2025-05-04")
async def test_Renoweb(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Aalborg', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Aalborg_139322', 'address_id': '139322',
                'kommunenavn': 'Aalborg', 'address': 'Boulevarden 13'}
            if not CI:
                address_list = await gc.get_address_list('9000', 'Boulevarden', '13')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('9000', 'Boulevarden', '')
                assert len(address_list) == 435
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('9000', 'Boulevarden', '1')
                assert len(address_list) == 21

            async def get_data(*args, **kwargs):
                return aalborg_data_gh
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Aalborg_gh', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time(FREEZE_TIME)
async def test_Odense(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Odense', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Odense_112970', 'address_id': '112970',
                'kommunenavn': 'Odense', 'address': 'Flakhaven 2'}
            if not CI:
                address_list = await gc.get_address_list('5000', 'Flakhaven', '2')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('5230', 'Læssøegade', '')
                assert len(address_list) == 162
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('5230', 'Læssøegade', '1')
                assert len(address_list) == 95

            async def get_data(*args, **kwargs):
                return odense_ics_data
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Odense', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time(FREEZE_TIME)
async def test_Aarhus(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Aarhus', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Aarhus_07517005___1__2____', 'address_id': '07517005___1__2____',
                'kommunenavn': 'Aarhus', 'address': 'Rådhuspladsen 1, 2.'}
            if not CI:
                address_list = await gc.get_address_list('8000', 'Rådhuspladsen', '2')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('8000', 'Rådhuspladsen', '')
                assert len(address_list) == 44
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('8000', 'Rådhuspladsen', '2')
                assert len(address_list) == 11

            async def get_data(*args, **kwargs):
                return aarhus_data[0]["plannedLoads"]
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Aarhus', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time("2025-05-18")
async def test_VestFor(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Ballerup', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Ballerup_2690c90b-016f-e511-80cd-005056be6a4c',
                'address_id': '2690c90b-016f-e511-80cd-005056be6a4c',
                'kommunenavn': 'Ballerup', 'address': 'Banegårdspladsen 1'}
            if not CI:
                address_list = await gc.get_address_list('2750', 'Banegårdspladsen', '1')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('2750', 'Banegårdspladsen', '')
                assert len(address_list) == 63
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('2750', 'Banegårdspladsen', '1')
                assert len(address_list) == 5

            async def get_data(*args, **kwargs):
                return vestfor_data
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Ballerup', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time("2025-06-04")
async def test_Provas(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Haderslev', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Haderslev_UHJvcGVydHlUeXBlOjEwMTQzNDE=', 'address_id': 'UHJvcGVydHlUeXBlOjEwMTQzNDE=', 
                'kommunenavn': 'Haderslev', 'address': "Christian x's vej 29"
                }
            if not CI:
                address_list = await gc.get_address_list('6100', "Christian X Vej", '29')
                address = await gc.get_address(address_list[0])
    #            print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('6100', 'Parkvej', '')
                assert len(address_list) == 53
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('6100', 'Parkvej', '2')
                assert len(address_list) == 9

            async def get_data(*args, **kwargs):
                return provas_data
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Haderslev', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time("2025-06-04")
async def test_RenoDjurs(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Norddjurs', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'Norddjurs_40130', 'address_id': '40130',
                'kommunenavn': 'Norddjurs', 'address': 'Torvet 3'
                }
            if not CI:
                address_list = await gc.get_address_list('8500', 'Torvet', '3')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('8500', 'Fuglsangparken', '')
                assert len(address_list) == 25
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('8500', 'Fuglsangparken', '1')
                assert len(address_list) == 16

            async def get_data(*args, **kwargs):
                return renodjurs_data
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Norddjurs', pickups, UPDATE)
            print('done: ', gc._municipality)


@pytest.mark.asyncio
@freeze_time(FREEZE_TIME)
async def test_Kbh(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('København', session=session, fail=True)
            print('start: ', gc._municipality)

            add = {
                'uid': 'København_a4e9a503-c27f-ef11-9169-005056823710',
                'address_id': 'a4e9a503-c27f-ef11-9169-005056823710',
                'kommunenavn': 'København', 'address': 'Rådhuspladsen 1'}
            if not CI:
                address_list = await gc.get_address_list('1550', 'Rådhuspladsen', '1')
                address = await gc.get_address(address_list[0])
                # print(address.__dict__)
                assert address.__dict__ == add
                address_list = await gc._api.get_address_list('2300', 'Irlandsvej', '')
                assert len(address_list) == 50
                await assert_add_list(gc, address_list)
                address_list = await gc._api.get_address_list('2300', 'Irlandsvej', '1')
                assert len(address_list) == 14

            async def get_data(*args, **kwargs):
                return kbh_ics_data
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data(add['address_id'])
            update_and_compare('Kbh', pickups, UPDATE)
            assert pickups['next_pickup'].description == 'Rest/Madaffald'
            assert pickups['next_pickup'].date.strftime('%d/%m/%y') == '05/05/25'
            assert list(pickups.keys()) == ['restaffaldmadaffald', 'farligtaffald', 'next_pickup']
            print('done: ', gc._municipality)



import pytest
from freezegun import freeze_time
from aiohttp import ClientSession
import datetime as dt
from pyaffalddk import GarbageCollection, NAME_LIST, api, const
from .data import const_tests
from pathlib import Path
import pickle
import json
from ical.calendar_stream import IcsCalendarStream


datadir = Path(__file__).parent/'data'


# always sort and overwrite the supported_items.json when running pytest
for key, vals in const.SUPPORTED_ITEMS.items():
    const.SUPPORTED_ITEMS[key] = sorted(list(set(vals)), key=str.lower)
with open(datadir.parents[1] / 'pyaffalddk/supported_items.json', 'w', encoding="utf-8") as fh:
    json.dump(const.SUPPORTED_ITEMS, fh, indent=4, ensure_ascii=False)


@pytest.mark.asyncio
@freeze_time("2025-05-09")
async def test_smoketest(capsys, monkeypatch, update=False):
    with capsys.disabled():
        async with ClientSession() as session:
            smokedata = pickle.load((datadir / 'smoketest_garbage_data.p').open('rb'))
            smokecompare_file = datadir / 'smoketest_fractions.json'
            with smokecompare_file.open('r') as fh:
                smokecompare = json.load(fh)

            for name, val in smokedata.items():
                city = val['city']
                gc = GarbageCollection(city, session=session, fail=True)

                async def get_data(*args, **kwargs):
                    return val['data']
                monkeypatch.setattr(gc._api, "get_garbage_data", get_data)
                pickups = await gc.get_pickup_data(1111)
                keys = list(pickups.keys())

                assert set(keys[:-1]).issubset(NAME_LIST)
                assert keys[-1] == 'next_pickup'

                data = {key: pickups[key].description for key in keys[:-1]}
                if name not in smokecompare or update:
                    smokecompare[name] = data
                    print(f'adding "{name}" to the smoketest compare data')
                    with smokecompare_file.open('w', encoding="utf-8") as fh:
                        json.dump(smokecompare, fh, indent=2, ensure_ascii=False)
                if smokecompare[name] != data:
                    print(name, city)
                    print(data)
                    print(smokecompare[name])
                assert smokecompare[name] == data


@pytest.mark.asyncio
@freeze_time("2025-05-22 12:30:00")
async def test_switch_next(capsys, monkeypatch):
    with capsys.disabled():
        async with ClientSession() as session:
            gc = GarbageCollection('Viborg', session=session, fail=True)
            openexp_data = json.loads((datadir/'openexp.data').read_text())

            async def get_data(*args, **kwargs):
                return openexp_data
            monkeypatch.setattr(gc._api, "get_garbage_data", get_data)

            pickups = await gc.get_pickup_data('1111')
            assert pickups['next_pickup'].date == dt.date(2025, 5, 22)

            gc.set_switch_time(12, 0)
            pickups = await gc.get_pickup_data('1111')
            assert pickups['next_pickup'].date == dt.date(2025, 6, 4)


def test_type_from_material_cleaning(capsys, monkeypatch):
    with capsys.disabled():
        for category, vals in const_tests.MATERIAL_LIST.items():
            for val in vals:
                cat = api.get_garbage_type(val, 'test', '1111', fail=False)
                if cat != category:
                    print(val, api.clean_fraction_string(val))
                assert cat == category

        for val in const_tests.NON_MATERIAL_LIST:
            cat = api.get_garbage_type(val, 'test', '1111', fail=False)
            assert cat == 'not-supported'


def test_type_cleaning(capsys, monkeypatch):
    with capsys.disabled():
        for category, vals in const_tests.SUPPORTED_ITEMS.items():
            for val in vals:
                cat = api.get_garbage_type(val, 'test', '1111', fail=False)
                if cat != category:
                    print(val, api.clean_fraction_string(val))
                assert cat == category


def test_ics(capsys):
    with capsys.disabled():
        odense_ics_data = (datadir/'odense_ics.data').read_text()
        ics = IcsCalendarStream.calendar_from_ics(odense_ics_data)
        data = odense_ics_data.replace("END:VTIMEZONE", """BEGIN:STANDARD
DTSTART:19701025T030000
RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=10
TZOFFSETFROM:+0200
TZOFFSETTO:+0100
END:STANDARD
BEGIN:DAYLIGHT
DTSTART:19700329T020000
RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=3
TZOFFSETFROM:+0100
TZOFFSETTO:+0200
END:DAYLIGHT
END:VTIMEZONE""")
        ics2 = IcsCalendarStream.calendar_from_ics(data)
        assert ics.events == ics2.events



"""This module contains the code to get garbage data from AffaldDK."""

import datetime as dt
from ical.calendar_stream import IcsCalendarStream
from ical.exceptions import CalendarParseError
import logging
import re
from typing import Any
import aiohttp
from dateutil import parser

from .const import (
    ICON_LIST,
    NAME_LIST,
    NON_SUPPORTED_ITEMS,
    RE_WORDS,
    RE_RAW,
    SPECIAL_MATERIALS,
    STRIPS,
    SUPPORTED_ITEMS,
    WEEKDAYS,
)
from .municipalities import MUNICIPALITIES_IDS, MUNICIPALITIES_LIST
from .data import PickupEvents, PickupType, AffaldDKAddressInfo
from pyaffalddk import interface

_LOGGER = logging.getLogger(__name__)


APIS = {
    'odense': interface.OdenseAffaldAPI,
    'aarhus': interface.AarhusAffaldAPI,
    'nemaffald': interface.NemAffaldAPI,
    'perfectwaste': interface.PerfectWasteAPI,
    'renoweb': interface.RenowebghAPI,
    'vestfor': interface.VestForAPI,
    'affaldonline': interface.AffaldOnlineAPI,
    'openexp': interface.OpenExperienceAPI,
    'openexplive': interface.OpenExperienceLiveAPI,
    'provas': interface.ProvasAPI,
    'renodjurs': interface.RenoDjursAPI,
}


class GarbageCollection:
    """Class to get garbage collection data."""

    def __init__(
        self,
        municipality: str,
        session: aiohttp.ClientSession = None,
        switch_hour: int = 15,
        fail: bool = False,
    ) -> None:
        """Initialize the class."""
        self._municipality = municipality
        self._street = None
        self._house_number = None
        self._api_type = None
        self._address_id = None
        self.fail = fail
        self.set_switch_time(switch_hour, 0)
        self.utc_offset = dt.datetime.now().astimezone().utcoffset()
        self.today = None
        self.last_fetch = ''
        municipality_id = MUNICIPALITIES_IDS.get(self._municipality.lower(), '')

        for key, value in MUNICIPALITIES_LIST.items():
            if key.lower() == self._municipality.lower():
                self._api_type = value[0]
                if len(value) > 1:
                    municipality_id = value[1]
                self._api = APIS[self._api_type](municipality_id, session=session)

        if self._api_type is None:
            raise RuntimeError(f'Unknow municipality: "{municipality}"')

    def set_switch_time(self, hours, minutes):
        self.switch_time = dt.time(hours, minutes)

    async def async_init(self) -> None:
        """Initialize the connection."""
        if self._municipality is not None:
            if self._api_type == "nemaffald":
                await self._api.token

    async def get_address_list(self, zipcode, street, house_number):
        """Get list of address, id """
        if self._api_type is not None:
            address_list = await self._api.get_address_list(zipcode, street, house_number)
            return sorted(address_list, key=extract_house_number)
        raise interface.AffaldDKNotSupportedError("Cannot find Municipality")

    async def get_address(self, address_name) -> AffaldDKAddressInfo:
        """Get the address from id"""

        if self._api_type is not None:
            self._address_id, address = await self._api.get_address(address_name)

            if self._address_id is None:
                raise interface.AffaldDKNotValidAddressError("Address not found")

            address_data = AffaldDKAddressInfo(
                str(self._address_id),
                self._municipality.capitalize(),
                address.capitalize()
            )
            return address_data
        else:
            raise interface.AffaldDKNotSupportedError("Cannot find Municipality")

    def update_pickup_event(self, item_name, address_id, _pickup_date):
        if _pickup_date is not None and _pickup_date < self.today:
            return 'old-event'

        key = get_garbage_type(item_name, self._municipality, address_id, self.fail)
        if key in ['not-supported', 'missing-type']:
            return key

        if (key not in self.pickup_events) or (_pickup_date < self.pickup_events[key].date):
            _pickup_event = {
                key: PickupType(
                    date=_pickup_date,
                    group=key,
                    friendly_name=NAME_LIST.get(key),
                    icon=ICON_LIST.get(key),
                    entity_picture=f"{key}.svg",
                    description=item_name,
                )
            }
            self.pickup_events.update(_pickup_event)
        return 'done'

    def set_next_event(self):
        if not self.pickup_events:
            return
        if dt.datetime.now().time() > self.switch_time:
            eventdates = [event.date for event in self.pickup_events.values() if event.date is not None and event.date > self.today]
        else:
            eventdates = [event.date for event in self.pickup_events.values() if event.date is not None]

        if eventdates:
            _next_pickup = min(eventdates)
            _next_name = [event.friendly_name for event in self.pickup_events.values() if event.date == _next_pickup]
            _next_description = [event.description for event in self.pickup_events.values() if event.date == _next_pickup]
            _next_pickup_event = {
                "next_pickup": PickupType(
                    date=_next_pickup,
                    group="genbrug",
                    friendly_name=list_to_string(_next_name),
                    icon=ICON_LIST.get("genbrug"),
                    entity_picture="genbrug.svg",
                    description=list_to_string(_next_description),
                )
            }
            self.pickup_events.update(_next_pickup_event)

    async def get_pickup_data(self, address_id: str, debug=False) -> PickupEvents:
        """Get the garbage collection data."""

        if (self._api_type is not None) & (self.today != dt.date.today()):
            self.pickup_events: PickupEvents = {}
            self.today = dt.date.today()
            self.last_fetch = str(dt.datetime.now())

            if self._api_type == 'odense':
                data = await self._api.get_garbage_data(address_id)
                try:
                    ics = IcsCalendarStream.calendar_from_ics(data)
                    for event in ics.timeline:
                        _garbage_types = split_ical_garbage_types(event.summary)
                        for garbage_type in _garbage_types:
                            _pickup_date = event.start_datetime.date()
                            self.update_pickup_event(garbage_type, address_id, _pickup_date)
                except CalendarParseError as err:
                    _LOGGER.error("Error parsing iCal data: %s", err)

            elif self._api_type == "aarhus":
                garbage_data = await self._api.get_garbage_data(address_id)
                for row in garbage_data:
                    _pickup_date = iso_string_to_date(row["date"])
                    for garbage_type in row["fractions"]:
                        self.update_pickup_event(garbage_type, address_id, _pickup_date)

            elif self._api_type == "nemaffald":
                data = await self._api.get_garbage_data(address_id)
                try:
                    ics = IcsCalendarStream.calendar_from_ics(data)
                    for event in ics.timeline:
                        _garbage_types = split_ical_garbage_types(event.summary)
                        for garbage_type in _garbage_types:
                            _pickup_date = (event.start_datetime + self.utc_offset).date()
                            self.update_pickup_event(garbage_type, address_id, _pickup_date)
                except CalendarParseError as err:
                    _LOGGER.error("Error parsing iCal data: %s", err)

            elif self._api_type == "perfectwaste":
                garbage_data = await self._api.get_garbage_data(address_id)
                for row in garbage_data:
                    _pickup_date = iso_string_to_date(row["date"])
                    for item in row["fractions"]:
                        garbage_type = item['fractionName']
                        self.update_pickup_event(garbage_type, address_id, _pickup_date)

            elif self._api_type == "renoweb":
                garbage_data = await self._api.get_garbage_data(address_id)
                for item in garbage_data:
                    if item['nextpickupdatetimestamp']:
                        _pickup_date = dt.datetime.fromtimestamp(int(item["nextpickupdatetimestamp"])).date()
                        self.update_pickup_event(item['name'], address_id, _pickup_date)

            elif self._api_type == "vestfor":
                garbage_data = await self._api.get_garbage_data(address_id)
                for item in garbage_data:
                    _pickup_date = iso_string_to_date(item['start'])
                    self.update_pickup_event(item['title'], address_id, _pickup_date)

            elif self._api_type == "affaldonline":
                garbage_data = await self._api.get_garbage_data(address_id)
                for item in garbage_data:
                    _pickup_date = iso_string_to_date(item['date'])
                    for garbage_type in item['collections']:
                        for container_description in set([cont['description'] for cont in garbage_type['containers']]):
                            self.update_pickup_event(container_description, address_id, _pickup_date)

            elif self._api_type == "openexp":
                garbage_data = await self._api.get_garbage_data(address_id)
                for item in garbage_data['collections']:
                    dt_list = [iso_string_to_date(d) for d in item['dates']]
                    if dt_list:
                        _pickup_date = min([d for d in dt_list if d >= self.today])
                        fraction_name = item['fraction']['name']
                        self.update_pickup_event(fraction_name, address_id, _pickup_date)

            elif self._api_type == "openexplive":
                garbage_data = await self._api.get_garbage_data(address_id)
                for item in garbage_data:
                    fraction_name = item['fraction']['name']
                    dt_list = [iso_string_to_date(d['date']) for d in item['upcoming_dates']]
                    if dt_list:
                        _pickup_date = min([d for d in dt_list if d >= self.today])
                        self.update_pickup_event(fraction_name, address_id, _pickup_date)

            elif self._api_type == "provas":
                garbage_data = await self._api.get_garbage_data(address_id)
                for item in garbage_data:
                    _pickup_date = iso_string_to_date(item['date'])
                    fraction_name = item['container']['waste_fraction']['name']
                    self.update_pickup_event(fraction_name, address_id, _pickup_date)

            elif self._api_type == "renodjurs":
                garbage_data = await self._api.get_garbage_data(address_id)
                for item in garbage_data:
                    _pickup_date = iso_string_to_date(item['Næste tømningsdag'], dayfirst=True)
                    fraction_name = item['Ordning']
                    self.update_pickup_event(fraction_name, address_id, _pickup_date)

        self.set_next_event()
        return self.pickup_events


def iso_string_to_date(datetext: str, dayfirst=None) -> dt.date:
    """Convert a date string to a datetime object."""
    if datetext == "Ingen tømningsdato fundet!":
        return None
    return parser.parse(datetext, dayfirst=dayfirst).date()


def get_garbage_type(item, municipality, address_id, fail=False):
    """Get the garbage type."""
    # _LOGGER.debug("Affalds type: %s", item)
    if item in NON_SUPPORTED_ITEMS:
        return 'not-supported'

    for special in SPECIAL_MATERIALS:
        if special.lower() in item.lower():
            return SPECIAL_MATERIALS[special]

    fixed_items = clean_fraction_string(item)
    for fixed_item in fixed_items:
        if fixed_item in [non.lower() for non in NON_SUPPORTED_ITEMS]:
            return 'not-supported'
        for key, values in SUPPORTED_ITEMS.items():
            for entry in values:
                if fixed_item.lower() == entry.lower():
                    return key
    print(f'\nmissing: {fixed_items}')
    warn_or_fail(item, municipality, address_id, fail=fail)
    return 'missing-type'


def clean_fraction_string(item):
    fixed_item = item.lower()

    for strip in WEEKDAYS + STRIPS:
        fixed_item = fixed_item.replace(strip.lower(), '')

    strings_in_parenthesis = re.findall(r'\(([^()]*)\)', fixed_item)
    pattern = r"\s*\([^()]*\)"
    fixed_item = re.sub(pattern, "", fixed_item)  # strip anything in parenthesis

    for word in RE_RAW:
        fixed_item = re.sub(word, '', fixed_item)
    for word in RE_WORDS:
        fixed_item = re.sub(fr'(?:\s|\b){word}(?=\s|$)', '', fixed_item)

    if ':' in fixed_item:
        fixed_item = fixed_item.split(':')[1]

    fixed_item = fixed_item.strip().rstrip(',').lstrip(', ').rstrip(' -').lstrip('- ').lstrip('*')
    res = [fixed_item.strip()]
    if ' - ' in fixed_item:
        res += [o.strip() for o in fixed_item.split(' - ')]
    return res + strings_in_parenthesis


def warn_or_fail(name, municipality, address_id, fail=False):
    msg = f'Garbage type [{name}] is not defined in the system. '
    msg += f'Please notify the developer. Municipality: {municipality}, Address ID: {address_id}'

    if fail:
        raise RuntimeError(msg)
    _LOGGER.warning(msg)


def list_to_string(list: list[str]) -> str:
    """Convert a list to a string."""
    return " | ".join(list)


def split_ical_garbage_types(text: str) -> list[str]:
    """Split a text string at every comma and ignore everything from 'på' or if it starts with 'Tømning af'."""
    if text.startswith("Tømning af"):
        text = text[len("Tømning af "):]
    if "på" in text:
        text = text.split("på")[0]
    return [item.strip() for item in text.split(",")]


def key_exists_in_pickup_events(pickup_events: PickupEvents, key: str) -> bool:
    """Check if a key exists in PickupEvents."""
    return key in pickup_events


def value_exists_in_pickup_events(pickup_events: PickupEvents, value: Any) -> bool:
    """Check if a value exists in PickupEvents."""
    return any(event for event in pickup_events.values() if event == value)


def extract_house_number(addr):
    """Sort address string by found house number numerical"""
    parts = addr.split(',', 1)
    second_part = parts[1].strip().lower() if len(parts) > 1 else ""
    match = re.search(r'\d+', parts[0])
    house_num = int(match.group()) if match else float('inf')  # non-numbered addresses go last
    return (house_num, second_part)



import aiohttp
import base64
import datetime as dt
import logging
import re
import json
from urllib.parse import urlparse, parse_qsl, quote
from bs4 import BeautifulSoup

from .const import GH_API


_LOGGER = logging.getLogger(__name__)


def clean_name(name):
    # removes everything from the last comma if postalcode in last part
    return re.sub(r',\s*\d{4}.*$', '', name)


class AffaldDKNotSupportedError(Exception):
    """Raised when the municipality is not supported."""


class AffaldDKNotValidAddressError(Exception):
    """Raised when the address is not found."""


class AffaldDKNoConnection(Exception):
    """Raised when no data is received."""


class AffaldDKGarbageTypeNotFound(Exception):
    """Raised when new garbage type is detected."""


class AffaldDKAPIBase:
    """Base class for the API."""

    def __init__(self, municipality_id, session=None) -> None:
        """Initialize the class."""
        self.municipality_id = municipality_id
        self.session = session
        self.address_list = {}
        if self.session is None:
            self.session = aiohttp.ClientSession()

    async def get_address(self, address_name):
        if address_name in self.address_list:
            return self.address_list[address_name]['id'], address_name
        return None, None

    def update_address_list(self, item, name_key, id_key):
        name = clean_name(item[name_key])
        if name in self.address_list:
            name += ' *'
            while name in self.address_list:
                name += '*'
        self.address_list[name] = {'id': item[id_key], 'fullname': item[name_key]}

    async def async_get_request(self, url, headers=None, para=None, as_json=True, new_session=False):
        return await self.async_api_request('GET', url, headers, para, as_json, new_session)

    async def async_post_request(self, url, headers={"Content-Type": "application/json"}, para=None, as_json=True, new_session=False):
        return await self.async_api_request('POST', url, headers, para, as_json, new_session)

    async def async_api_request(self, method, url, headers, para=None, as_json=True, new_session=False):
        """Make an API request."""

        if new_session:
            session = aiohttp.ClientSession()
        else:
            session = self.session

        data = None
        if method == 'POST':
            json_input = para
            data_input = None
        else:
            json_input = None
            data_input = para

        async with session.request(method, url, headers=headers, json=json_input, params=data_input) as response:
            if response.status != 200:
                if new_session:
                    await session.close()

                if response.status == 400:
                    raise AffaldDKNotSupportedError(
                        "Municipality not supported")

                if response.status == 404:
                    raise AffaldDKNotSupportedError(
                        "Municipality not supported")

                if response.status == 503:
                    raise AffaldDKNoConnection(
                        "System API is currently not available")

                raise AffaldDKNoConnection(
                    f"Error {response.status} from {url}")
            if as_json:
                data = await response.json()
            else:
                data = await response.text()
            if new_session:
                await session.close()

            return data


class NemAffaldAPI(AffaldDKAPIBase):
    # NemAffaldService API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._token = None
        self._id = None
        self.street = None
        self.base_url = f'https://nemaffaldsservice.{self.municipality_id}.dk'

    @property
    async def token(self):
        if self._token is None:
            await self._get_token()
        return self._token

    async def _get_token(self):
        data = ''
        async with self.session.get(self.base_url) as response:
            data = await response.text()

        if data:
            match = re.search(
                r'name="__RequestVerificationToken"\s+[^>]*value="([^"]+)"', data)
            if match:
                self._token = match.group(1)

    async def get_address_list(self, zipcode, street, house_number):
        url = self.base_url + '/WasteHome/AddressByTerm'
        para = {'term': f'{street} {house_number}'.strip(), 'limit': 100}
        data = await self.async_get_request(url, para=para, as_json=False)
        data = json.loads(data)
        self.address_list = {}
        for item in data:
            self.update_address_list(item, 'label', 'Id')
        return list(self.address_list.keys())

    async def get_address(self, address_name):
        url = f"{self.base_url}/WasteHome/SearchCustomerRelation"
        data = {'__RequestVerificationToken': await self.token, 'SearchTerm': address_name}
        async with self.session.post(url, data=data) as response:
            if len(response.history) > 1:
                o = urlparse(response.history[1].headers['Location'])
                id = dict(parse_qsl(o.query))['customerId']
                return id, address_name
        return None, None

    async def async_get_ical_data(self, customerid):
        ics_data = ''
        data = {'customerId': customerid, 'type': 'ics'}
        async with self.session.get(f"{self.base_url}/Calendar/GetICaldendar", data=data) as response:
            ics_data = await response.text()
        return ics_data

    async def get_garbage_data(self, address_id):
        return await self.async_get_ical_data(address_id)


class VestForAPI(AffaldDKAPIBase):
    # Vest Forbrænding API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.baseurl = "https://selvbetjening.vestfor.dk"
        self.url_data = self.baseurl + "/Home/MinSide"
        self.url_search = self.baseurl + "/Adresse/AddressByName"
        self.today = dt.date.today()

    async def get_address_list(self, zipcode, street, house_number):
        para = {'term': f'{street} {house_number}'.strip(), 'numberOfResults': 100}
        data = await self.async_get_request(self.url_search, para=para)
        self.address_list = {}
        for item in data:
            if str(zipcode) in item['Postnr']:
                self.update_address_list(item, 'FuldtVejnavn', 'Id')
        return list(self.address_list.keys())

    async def get_garbage_data(self, address_id):
        para = {'address-selected-id': address_id}
        _ = await self.async_get_request(self.url_data, para=para, as_json=False)
        url = 'https://selvbetjening.vestfor.dk/Adresse/ToemmeDates'
        para = {
            'start': str(self.today + dt.timedelta(days=-1)),
            'end': str(self.today + dt.timedelta(days=60))
            }
        data = await self.async_get_request(url, para=para)
        return data


class PerfectWasteAPI(AffaldDKAPIBase):
    # Perfect Waste API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.baseurl = "https://europe-west3-perfect-waste.cloudfunctions.net"
        self.url_data = self.baseurl + "/getAddressCollections"
        self.url_search = self.baseurl + "/searchExternalAddresses"

    async def get_address_list(self, zipcode, street, house_number):
        para = {'data': {
            "query": f"{street} {house_number}".strip(),
            "municipality": self.municipality_id,
            "page": 1, "onlyManual": False
            }}

        page = await self.async_post_request(self.url_search, para=para)
        data = page['result']
        while len(page['result']) == 30:
            para['data']['page'] += 1
            page = await self.async_post_request(self.url_search, para=para)
            data += page['result']
        self.address_list = {}
        for item in data:
            if str(zipcode) in item['displayName']:
                self.update_address_list(item, 'displayName', 'addressID')
        return list(self.address_list.keys())

    async def save_to_db(self, address_id):
        url = self.baseurl + '/fetchAddressAndSaveToDb'
        para = {"data": {
            "addressID": address_id, "municipality": self.municipality_id,
            "caretakerCode": None, "isCaretaker": None}}
        await self.async_post_request(url, para=para)

    async def get_garbage_data(self, address_id):
        body = {"data": {
            "addressID": address_id,
            "municipality": self.municipality_id
            }}
        data = await self.async_post_request(self.url_data, para=body)
        return data["result"]


class RenowebghAPI(AffaldDKAPIBase):
    # Renoweb servicegh API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_data = "https://servicesgh.renoweb.dk/v1_13/"
        self.uuid = base64.b64decode(GH_API).decode('utf-8')
        self.headers = {'Accept-Encoding': 'gzip'}

    async def get_address_list(self, zipcode, street, house_number):
        road_id = await self.get_road(zipcode, street)
        if road_id:
            url = self.url_data + 'GetJSONAdress.aspx'
            para = {'apikey': self.uuid, 'municipalitycode': self.municipality_id, 'roadid': road_id}
            if house_number:
                para['streetBuildingIdentifier'] = house_number
            data = await self.async_get_request(url, para=para, headers=self.headers)
            self.address_list = {}
            for item in data['list']:
                if str(zipcode) in item['presentationString']:
                    self.update_address_list(item, 'presentationString', 'id')
            return list(self.address_list.keys())
        return []

    async def get_road(self, zipcode, street):
        url = self.url_data + 'GetJSONRoad.aspx'
        data = {
            'apikey': self.uuid, 'municipalitycode': self.municipality_id,
            'roadname': street
        }
        js = await self.async_get_request(url, para=data, headers=self.headers)
        for item in js['list']:
            if str(zipcode) in item['name']:
                return item['id']
        return None

    async def get_garbage_data(self, address_id, fullinfo=0, shared=0):
        url = self.url_data + 'GetJSONContainerList.aspx'
        data = {
            'apikey': self.uuid, 'municipalitycode': self.municipality_id,
            'adressId': address_id, 'fullinfo': fullinfo, 'supportsSharedEquipment': shared,
            }
        js = await self.async_get_request(url, para=data, headers=self.headers)
        if js:
            return js['list']
        return []


class AffaldOnlineAPI(AffaldDKAPIBase):
    # Affald online / Renoweb API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_base = "https://www.affaldonline.dk/api/address/"
        uuid = base64.b64decode(self.municipality_id).decode('utf-8')
        self.headers = {
            'X-Client-Provider': uuid,
            'X-Client-Type': 'Kunde app',
            'X-Client-Version': '22',
        }

    async def get_address_list(self, zipcode, street, house_number):
        para = {'q': f'{street} {house_number}'.strip(), 'page': 1}
        page = await self.async_get_request(self.url_base + 'search', para=para, headers=self.headers)
        data = page['results']
        while len(page['results']) == 50:
            para['page'] += 1
            page = await self.async_get_request(self.url_base + 'search', para=para, headers=self.headers)
            data += page['results']
        self.address_list = {}
        for item in data:
            if str(zipcode) in item['displayName']:
                self.update_address_list(item, 'displayName', 'addressId')
        return list(self.address_list.keys())

    async def get_garbage_data(self, address_id):
        para = {'groupBy': 'date', 'addressId': address_id}
        data = await self.async_get_request(self.url_base + 'collections', para=para, headers=self.headers)
        return data


class OpenExperienceAPI(AffaldDKAPIBase):
    # Open Experience API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_base = f"https://{self.municipality_id}"

    async def get_address_list(self, zipcode, street, house_number):
        address = f'{street} {house_number}'.strip()
        url = self.url_base + f'search/address/{quote(address)}/limit/200'
        data = await self.async_get_request(url)
        self.address_list = {}
        for item in sorted(data['results'], key=lambda x: int(x["addressId"])):
            if str(zipcode) in item['displayName']:
                self.update_address_list(item, 'displayName', 'addressId')
        return list(self.address_list.keys())

    async def get_garbage_data(self, address_id):
        url = self.url_base + f'address/{address_id}/collections'
        data = await self.async_get_request(url)
        return data


class OpenExperienceLiveAPI(AffaldDKAPIBase):
    # Open Experience API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_base = "https://live.affaldsapi.dk/"
        uuid = base64.b64decode(self.municipality_id[1:]).decode('utf-8')
        self.headers = {'X-Auth-Token': uuid}
        self.mid = self.municipality_id[0]

    async def get_address_list(self, zipcode, street, house_number):
        address = f'{street} {house_number}'.strip()
        url = self.url_base + f'address/v1/search/{self.mid}/{quote(address)}/100'
        data = await self.async_get_request(url, headers=self.headers)
        self.address_list = {}
        for item in sorted(data, key=lambda x: int(x["id"])):
            if str(zipcode) in item['name']:
                self.update_address_list(item, 'name', 'id')
        return list(self.address_list.keys())

    async def get_garbage_data(self, address_id):
        url = self.url_base + f'arrangements/v1/collections/calendar/{self.mid}/{address_id}'
        data = await self.async_get_request(url, headers=self.headers)
        return data


class ProvasAPI(AffaldDKAPIBase):
    # Provas API (Haderslev)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_base = "https://platform-api.wastehero.io/api-crm-portal/v1/"
        self._token = None
        self.today = dt.date.today()

    @property
    async def token(self):
        if not self._token:
            vars = {"username": "", "password": ""}
            url = self.url_base + 'company/provas-portal.wastehero.io/login'
            data = await self.async_post_request(url, para=vars)
            self._token = data.get('token')
        return self._token

    async def get_address_list(self, zipcode, street, house_number):
        headers = {'X-API-Key': await self.token}
        address = f'{street} {house_number}, {zipcode}'.strip()
        url = self.url_base + 'property/'
        data = await self.async_get_request(url, para={'search': address, 'limit':100}, headers=headers)
        self.address_list = {}
        for item in data:
            item['name'] = item['location']['name']
            if str(zipcode) in item['name']:
                self.update_address_list(item, 'name', 'id')
        return list(self.address_list.keys())

    async def get_garbage_data(self, address_id):
        headers = {'X-API-Key': await self.token}
        url = self.url_base + f'property/{address_id}/collection_log'
        params = {
            'from_date': str(self.today + dt.timedelta(days=-1)),
            'to_date': str(self.today + dt.timedelta(days=60))
        }
        data = await self.async_get_request(url, para=params, headers=headers)
        return data


class RenoDjursAPI(AffaldDKAPIBase):
    # Reno Djurs API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_base = "https://minside.renodjurs.dk"

    async def get_address_list(self, zipcode, street, house_number):
        address = f'{street} {house_number}*, {zipcode}'.strip()
        url = self.url_base + '/Default.aspx/GetAddress'
        data = await self.async_post_request(url, para={'address': address.strip()})
        self.address_list = {}
        for item in data['d']:
            if str(zipcode) in item['label']:
                self.update_address_list(item, 'label', 'value')
        return list(self.address_list.keys())

    async def get_garbage_data(self, address_id):
        url = self.url_base + '/Ordninger.aspx?id=55424'
        data = await self.async_get_request(url, as_json=False)
        soup = BeautifulSoup(data, "html.parser")
        table = soup.find("table", class_="table--compact")

        headers = [th.get_text(strip=True) for th in table.find("thead").find_all("th")]
        rows = []
        for tr in table.find("tbody").find_all("tr"):
            td = tr.find_all("td")
            if len(td) == len(headers):
                row_data = {h:td.get_text(strip=True) for h, td in zip(headers, td)}
                if row_data['Næste tømningsdag']:
                    rows.append(row_data)
        return rows


class AarhusAffaldAPI(AffaldDKAPIBase):
    # Aarhus Forsyning API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_data = "https://portal-api.kredslob.dk/api/calendar/address/"
        self.url_search = "https://api.dataforsyningen.dk/adresser"

    async def get_address_list(self, zipcode, street, house_number):
        para = {'kommunekode': 751, 'q': f'{street} {house_number}'.strip(), 'struktur': 'mini'}
        data = await self.async_get_request(self.url_search, para=para)
        self.address_list = {}
        for item in data:
            if str(zipcode) in item['postnr']:
                self.update_address_list(item, 'betegnelse', 'id')
        return list(self.address_list.keys())

    async def get_address(self, address_name):
        item = self.address_list.get(address_name)
        if item:
            para = {'kommunekode': 751, 'id': item['id']}
            row = await self.async_get_request(self.url_search, para=para)
            return row[0]["kvhx"], address_name

    async def get_garbage_data(self, address_id):
        url = f"{self.url_data}{address_id}"
        data = await self.async_get_request(url)
        return data[0]["plannedLoads"]


class OdenseAffaldAPI(AffaldDKAPIBase):
    # Odense Renovation API
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url_data = "https://mit.odenserenovation.dk/api/Calendar/GetICalCalendar?addressNo="
        self.url_search = "https://mit.odenserenovation.dk/api/Calendar/CommunicationHouseNumbers?addressString="

    async def get_address_list(self, zipcode, street, house_number):
        address = f'{street} {house_number}'.strip()
        url = f"{self.url_search}{quote(address)}"
        data = await self.async_get_request(url)
        self.address_list = {}
        for item in data:
            if str(zipcode) in item['PostCode']:
                self.update_address_list(item, 'FullAddress', 'AddressNo')
        return list(self.address_list.keys())

    async def async_get_ical_data(self, address_id):
        """Get data from iCal API."""
        url = f"{self.url_data}{address_id}"
        data = await self.async_get_request(url, as_json=False)
        return data

    async def get_garbage_data(self, address_id):
        return await self.async_get_ical_data(address_id)
